import re
import zipfile
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import IO, TypedDict
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

import yaml

from anyerplint import preprocessor
from anyerplint.translate import Translator

from . import __version__, tnex, util

ROOT = Path(__file__).absolute().parent


ReportData = TypedDict(
    "ReportData",
    {
        "Unknown variables": list[str],
        "Unknown calls": list[str],
        "Unknown jumps": list[str],
        "Unknown tags": list[str],
        "Other errors": list[str],
        "Expression errors": list[dict[str, str | list[str]]],
        "Unknown loop variables": list[str],
    },
    total=False,
)


@dataclass
class Library:
    """Stores known symbols from the 'standard library'."""

    variables: set[str]
    loopvariables: set[str]
    calls: set[str]
    jumps: set[str]
    tags: set[str]
    # if true, accumulate everything found to library
    teaching: bool


# ruff lead this alone man
emit = print


def do_check(
    libs: list[str],
    targets: list[str],
    teaching: bool,
    *,
    nostdlib: bool = False,
) -> dict[str, str | ReportData]:
    emit(f"AnyErpLintVersion: {__version__.strip()}")

    lib_vars = Library(
        variables=set(),
        calls=set(),
        jumps=set(),
        tags=set(),
        teaching=teaching,
        loopvariables=set(),
    )

    # always feed the "standard library, except when unit testing (nostdlib)

    if not nostdlib:
        local_app_data = util.get_app_local_dir()
        feed_lib(lib_vars, local_app_data)

    for lib in libs:
        feed_lib(lib_vars, Path(lib))

    has_errors = False
    all_errors: dict[str, str | ReportData] = {}
    for target in targets:
        files = util.glob_plus(target)
        errs = check_files(lib_vars, files)
        if errs:
            all_errors.update(errs)
            has_errors = True

    if lib_vars.teaching:
        write_lib(lib_vars)

    if has_errors:
        emit("Errors found: >")
        error_counts = {
            f: (
                sum(len(prop) for prop in e.values() if isinstance(prop, list))
                if not isinstance(e, str)
                else -1
            )
            for f, e in all_errors.items()
        }
        rep = sorted((k, v) for (k, v) in error_counts.items())
        for line in rep:
            emit("  ", line[0], ";", line[1])

    return all_errors


def write_lib(lib_vars: Library) -> None:
    def write_file(fname: Path, lines: list[str]) -> None:
        emit("  - ", fname)
        with fname.open("wb") as f:
            for line in lines:
                try:
                    if not line.strip().isascii():
                        continue
                    enc = line.strip().encode()
                except UnicodeEncodeError:
                    # skip bad lines for now
                    continue
                f.write(enc)
                f.write(b"\n")

    emit("Writing found function to:")

    stdlib = util.get_app_local_dir()
    calls_file = stdlib / "builtin_calls.txt"
    write_file(
        calls_file,
        sorted(call for call in lib_vars.calls if is_valid_call(call)),
    )

    jumps_file = stdlib / "builtin_jumps.txt"
    write_file(jumps_file, sorted(lib_vars.jumps))

    tags_file = stdlib / "builtin_tags.txt"
    write_file(tags_file, sorted(lib_vars.tags))

    vars_file = stdlib / "builtin_vars.txt"
    write_file(vars_file, sorted(lib_vars.variables))

    loopvars_file = stdlib / "builtin_loopvars.txt"
    write_file(loopvars_file, sorted({lv.lower() for lv in lib_vars.loopvariables}))


def is_valid_call(_call: str) -> bool:
    return True


def feed_lib(lib_vars: Library, libdir: Path) -> None:
    def feed_set(dest: set[str], fobj: IO[bytes]) -> None:
        dest.update(line.decode().strip() for line in fobj.readlines())

    def visit_file(fname: str, fobj: IO[bytes]) -> None:
        if fname.endswith("_calls.txt"):
            feed_set(lib_vars.calls, fobj)
        elif fname.endswith("_jumps.txt"):
            feed_set(lib_vars.jumps, fobj)
        elif fname.endswith("_tags.txt"):
            feed_set(lib_vars.tags, fobj)
        elif fname.endswith("_vars.txt"):
            feed_set(lib_vars.variables, fobj)
        elif fname.endswith("_loopvars.txt"):
            feed_set(lib_vars.loopvariables, fobj)
            lib_vars.loopvariables = {lv.lower() for lv in lib_vars.loopvariables}

    if not libdir.exists():
        return

    # files on file system
    for p in libdir.glob("*_*.txt"):
        visit_file(str(p), p.open("rb"))

    # files in all the zip files
    for f in libdir.glob("*.zip"):
        zf = zipfile.ZipFile(f, "r")
        for zn in zf.namelist():
            visit_file(zn, zf.open(zn, "r"))


def report_fatal(fname: Path, ex: Exception) -> str:
    message = f"FATAL, {ex}"
    emit(f"{fname}: {message}")
    return message


def should_skip_file(fpath: Path) -> bool:
    if fpath.suffix != ".xml":
        return True
    text = util.read_text_file(fpath)
    return not ("<erpConnector" in text or "<section" in text)


def check_file(lib_vars: Library, file: Path) -> str | ReportData | None:
    try:
        if should_skip_file(file):
            emit(f"{file}: SKIP nontemplate")
            return None
        r = parse_file(file, teaching=lib_vars.teaching)
    except (
        ElementTree.ParseError,
        PermissionError,
        UnicodeDecodeError,
        preprocessor.PreprocessError,
    ) as e:
        return report_fatal(file, e)

    return report(lib_vars, str(file), r)


def check_files(
    lib_vars: Library, files: Iterable[Path]
) -> dict[str, str | ReportData]:
    errs: dict[str, str | ReportData] = {}
    for f in files:
        errlist = check_file(lib_vars, f)
        if errlist:
            errs[str(f)] = errlist

    return errs


@dataclass
class ExprError:
    """Represents an error in an expression."""

    summary: str
    translated: str
    messages: list[str]

    def render(self) -> dict[str, str | list[str]]:
        """Render yaml compatible object."""
        return {"Error": self.summary, "Simple": self.translated, "Msg": self.messages}


@dataclass
class Parsed:
    """The things found while parsing a file."""

    var_decl: set[str]
    var_used: set[str]
    alltags: set[str]
    calls: set[str]
    jumps: set[str]
    syntax_errors: list[str]
    loop_var_decl: set[str]
    loop_var_use: set[str]
    expression_errors: list[ExprError]
    full_content: str


def report(lib_vars: Library, fname: str, p: Parsed) -> ReportData:
    undeclared_vars = p.var_used - p.var_decl
    undeclared_vars.difference_update(lib_vars.variables)
    unknown_loop_variables = {
        lv
        for lv in p.loop_var_use
        if lv.lower() not in p.loop_var_decl
        and lv.lower() not in lib_vars.loopvariables
    }

    if lib_vars.teaching:
        lib_vars.calls.update(p.calls)
        lib_vars.jumps.update(p.jumps)
        lib_vars.tags.update(p.alltags)
        lib_vars.variables.update(undeclared_vars)
        lib_vars.loopvariables.update(lv.lower() for lv in unknown_loop_variables)

    errors: ReportData = {}

    if undeclared_vars:
        errors["Unknown variables"] = util.add_linenumbers(
            p.full_content,
            sorted(undeclared_vars),
        )

    unknown_calls = p.calls
    unknown_calls.difference_update(lib_vars.calls)
    if unknown_calls:
        errors["Unknown calls"] = sorted(unknown_calls)

    unknown_jumps = p.jumps
    unknown_jumps.difference_update(lib_vars.jumps)
    if unknown_jumps:
        errors["Unknown jumps"] = sorted(unknown_jumps)

    unknown_tags = p.alltags
    unknown_tags.difference_update(lib_vars.tags)
    if unknown_tags:
        errors["Unknown tags"] = sorted(unknown_tags)

    if p.syntax_errors:
        errors["Other errors"] = list(p.syntax_errors)

    if p.expression_errors:
        errors["Expression errors"] = [e.render() for e in p.expression_errors]

    if unknown_loop_variables:
        errors["Unknown loop variables"] = sorted(unknown_loop_variables)

    if errors:
        emit(yaml.dump({fname: errors}, width=200, sort_keys=False).strip())

    return errors


key_params: dict[str, str] = {
    "bw_file_functions": "command",
    "bw_table_method": "command",
    "bw_string_functions": "operation",
    "bw_ws_function": "method",
}


def summarize_call(node: Element) -> str:
    name = node.attrib.get("name", "NONAME").lower()
    full = name
    # What should we do when there's multiple params with the same name?
    params = {
        p.attrib.get("name", "NONAME"): (p.text or "TEXTMISSING")
        for p in node.iter("parameter")
    }
    suboperation_param_name = key_params.get(name)
    if suboperation_param_name:
        suboperation = params.get(suboperation_param_name, "UNK")
        full += "." + suboperation

    return (full + " - " + ",".join(sorted(params))).strip()


def summarize_output(node: Element) -> str:
    output_type = node.attrib.get("type", "NOTYPE")
    params = {p.attrib.get("name", "NONAME") for p in node.iter("parameter")}

    return ("Output " + output_type + " - " + ",".join(sorted(params))).strip()


def summarize_tag(node: Element) -> str:
    at = " " + " ".join(sorted(node.attrib.keys())) if node.attrib else ""
    return "<" + node.tag + at + ">"


def brace_check(s: str) -> list[str]:
    stack: list[tuple[str, int]] = []
    lines = s.splitlines()
    closers = {"{": "}", "[": "]", "(": ")"}
    errors: list[str] = []
    for lnum, line in enumerate(lines, 1):
        flush_stack = False
        in_quote = False
        for cnum, ch in enumerate(line, 1):
            if ch == '"':
                # only care about quotes if we are in some nested operation already, top level quotes are not considered
                if stack:
                    in_quote = not in_quote

            if in_quote:
                continue

            if ch in "{([":
                stack.append((ch, lnum))
            if ch in "})]":
                try:
                    from_stack, _ = stack.pop()
                except IndexError:
                    errors.append(
                        f"Too many closing braces at line {lnum}, looking at '{ch}' on col {cnum}: ==> {line[cnum-10:cnum]} <==: {line.strip()}",
                    )
                    flush_stack = True
                    break

                expected = closers[from_stack]
                if expected != ch:
                    errors.append(
                        f"Expected brace {expected}, got {ch} at line {lnum} col {cnum}: {line.strip()}",
                    )
                    flush_stack = True
                    break
        if flush_stack:
            stack = []
    if stack:
        pretty_stack = ", ".join(f"{ch} {lnum}" for ch, lnum in stack)
        errors.append(
            f"File ended with mismatched braces, remaining in stack (char, linenum): {pretty_stack}",
        )
    return errors


def describe_node(n: Element) -> str:
    return "<" + n.tag + str(n.attrib) + ">"


def describe_jump(n: Element) -> str:
    params = sorted(
        child.attrib.get("name", "NONAME").strip() for child in n.iter("parameter")
    )
    target = n.attrib.get("jumpToXPath", "NOXPATH")
    prefix = "//section[@name='"
    if target.startswith(prefix):
        target = "..." + target[len(prefix) :].rstrip("]'")

    desc = (
        "Jump "
        + n.attrib.get("jumpToXmlFile", "NOFILE")
        + " -- "
        + target
        + " -- "
        + " ".join(params)
    )
    return desc.strip()


def is_illegal_password(name: str, value: str) -> bool:
    if "passw" not in name.lower():
        return False
    stripped = (value or "").strip()
    if not stripped:
        return False
    if stripped.startswith("{"):
        # password should always be references to variables or expressions, never literal values
        return False
    return True


XMLWRITER_COMMAND_PAIRS = {
    "startattribute": "endattribute",
    "startdocument": "enddocument",
    "startelement": "endelement",
}

XMLWRITER_COMMAND_SINGLE = {"write", "writeraw"}


def check_single_outputresource(elem: Element, output_location: str) -> str | None:
    cmd_stack = []
    for ind, cmd in enumerate(elem.iter("command"), 1):
        cmd_type = cmd.attrib.get("type")
        if not cmd_type:
            return f"Command missing 'type' attribute in OutputResource ({output_location}, command {ind})"
        elif cmd_type in XMLWRITER_COMMAND_SINGLE:
            continue
        elif cmd_type in XMLWRITER_COMMAND_PAIRS:
            cmd_stack.append(cmd_type)
        elif cmd_type in XMLWRITER_COMMAND_PAIRS.values():
            try:
                prev = cmd_stack.pop()
            except IndexError:
                return f"Invalid command '{cmd_type}' in OutputResource, no matching starting tag ({output_location}, command {ind})"
            expected = XMLWRITER_COMMAND_PAIRS.get(prev)
            if cmd_type != expected:
                return f"Invalid command type '{cmd_type}' in OutputResource, expected '{expected}' ({output_location}, command {ind})"
        else:
            return f"Unknown command type '{cmd_type}' in OutputResource ({output_location}, command {ind})"

    if cmd_stack:
        return f"Did not end all started items in OutputResource, unterminated: {','.join(cmd_stack)} ({output_location})"
    return None


def check_outputresources(root: Element) -> list[str]:
    errors = []
    for outputresource in root.iter("output"):
        if outputresource.get("type", "").split(".")[-1] == "XmlWriterOutputResource":
            description = util.format_xml_tag(outputresource)
            cmd_error = check_single_outputresource(outputresource, description)
            if cmd_error:
                errors.append(cmd_error)
    return errors


def full_parse_expression(expr: str) -> tuple[str, list[str]]:
    try:
        parsed = tnex.parse(expr, expand_entities=True)
        if not parsed:
            return "", [f"Empty parse result from: {expr}"]
        translator = Translator()
        translated = translator.translate(parsed)
    except tnex.ParseError as e:
        return "", [f"Fatal parse error: {e}"]
    else:
        return translated, translator.errors


MAX_EXPRESSION_LINES = 500


def find_invalid_expressions(expressions: list[tuple[str, int]]) -> list[ExprError]:
    res: list[ExprError] = []
    for expr, linenum in expressions:
        if expr and expr[0].isalpha() and "," not in expr:
            res.append(
                ExprError(
                    summary=f"Line {linenum}, expression contains no comma: {{{expr}}}",
                    translated="N/A",
                    messages=[],
                )
            )
            continue
        if expr.count("\n") > MAX_EXPRESSION_LINES:
            res.append(
                ExprError(
                    summary=f"Line {linenum}, expression too long, likely unterminated: {{{expr[:200]}}}",
                    translated="N/A",
                    messages=[],
                )
            )
            continue

        translated, errors = full_parse_expression(expr)

        if errors:
            res.append(
                ExprError(
                    summary=f"Line {linenum}, expression: {{{expr}}}",
                    translated=translated,
                    messages=errors,
                )
            )
    return res


def parse_file(fname: Path, *, teaching: bool = False) -> Parsed:
    tree = ElementTree.parse(fname)
    raw_cont = util.read_text_file(fname)
    comments_removed = util.replace_commented_xml_with_empty_lines(raw_cont)

    vardecl = {
        v.attrib.get("name", "unknown_var"): (v.text or "")
        for v in tree.iter("variable")
    }
    all_params = {
        v.attrib.get("name", "unknown_var"): (v.text or "")
        for v in tree.iter("parameter")
    }

    expressions = preprocessor.get_expressions(comments_removed)

    prop_regex = r"([a-zA-Z_][a-zA-Z0-9_.]*),(\w+)"

    propaccess = {
        (match.group(1), match.group(2))
        for expr, line in expressions
        for match in re.finditer(prop_regex, expr)
    }
    varuse = {name for expr_type, name in propaccess if expr_type.lower() == "v"}

    # what to do with p params?
    otherpropaccess = {k for k, v in propaccess if k.lower() not in ["v", "f", "p"]}
    calls = {summarize_call(v) for v in tree.iter("builtInMethodParameterList")}
    outputs = {summarize_output(v) for v in tree.iter("output")}
    calls.update(outputs)
    alltags = {summarize_tag(t) for t in tree.iter()}
    loop_data_source_attribs = {n.attrib.get("loopDataSource") for n in tree.iter()}
    loop_data_sources = {
        ls.split(";")[0].lower() for ls in loop_data_source_attribs if ls
    }
    return_names = {
        n.attrib.get("name", "UNNAMED_RETURN").lower() for n in tree.iter("return")
    }
    loop_data_sources.update(return_names)

    jumps = {
        describe_jump(n) for n in tree.iter("method") if n.attrib.get("jumpToXmlFile")
    }

    errors = []
    expr_errors = []
    if not teaching:
        cdata_removed = util.replace_cdata_with_empty_lines(comments_removed)
        errors.extend(brace_check(cdata_removed))

        no_text_allowed_tags = [
            "sections",
            "section",
            "method",
            "output",
            "outputCommands",
            "builtInMethodParameterList",
        ]
        for notext in no_text_allowed_tags:
            nodes = tree.iter(notext)
            for n in nodes:
                if n and n.text and n.text.strip():
                    errors.append(
                        "Node should not contain text: "
                        + describe_node(n)
                        + " -- "
                        + n.text.strip(),
                    )
        var_passwords = {v for v in vardecl if is_illegal_password(v, vardecl[v])}
        param_passwords = {
            p for p in all_params if is_illegal_password(p, all_params[p])
        }
        passwords = var_passwords | param_passwords
        if passwords:
            errors.append("Passwords contains literal text: " + ",".join(passwords))

        outputresource_errors = check_outputresources(tree.getroot())
        errors.extend(outputresource_errors)

        expr_errors = find_invalid_expressions(expressions)

    return Parsed(
        var_decl=set(vardecl),
        var_used=varuse,
        alltags=alltags,
        calls=calls,
        jumps=jumps,
        syntax_errors=errors,
        loop_var_decl=loop_data_sources,
        loop_var_use=otherpropaccess,
        full_content=comments_removed,
        expression_errors=expr_errors,
    )



import re
from collections.abc import Iterable

from anyerplint import util


class PreprocessError(Exception):
    """Raised by preprocessor."""


def tokenize(cont: str) -> list[str]:
    return re.split(r"([{}\"])", cont)


def walk_tokens(toks: list[str]) -> Iterable[tuple[str, int]]:
    in_expression = False
    in_string = False
    escaping = False
    acc = []
    current_start_line = 0
    lnum = 1
    str_tok_start_index = None
    expr_start_index = 0
    diag_strings: list[str] = []
    for idx, tok in enumerate(toks):
        if not tok:
            # skip ''
            continue
        if in_expression:
            acc.append(tok)

        match tok:
            case "{" if not in_string:
                if in_expression:
                    msg = f"Invalid character '{{' in expression, line {current_start_line}"
                    raise PreprocessError(msg)
                in_expression = True
                current_start_line = lnum
                expr_start_index = idx
            case "}" if not in_string:
                if not in_expression:
                    continue
                in_expression = False
                # pop already pushed }
                acc.pop()
                yield ("".join(acc), current_start_line)
                diag_strings = []
                acc = []
            case '"' if in_expression and not escaping:
                in_string = not in_string
                if in_string:
                    str_tok_start_index = idx
                else:
                    diag_strings.append("".join(toks[str_tok_start_index : idx + 1]))
            case _:
                ending_quote_count = util.count_trailing(tok, "\\")
                escaping = ending_quote_count % 2 != 0

        lnum += tok.count("\n")

    if in_expression:
        raise PreprocessError(
            f"Unterminated expression (too many {{ characters) at {current_start_line}. Starting with: "
            + repr("".join(toks[expr_start_index : expr_start_index + 10]))
        )
    if in_string:
        stringlist = " :: ".join(repr(s)[:500] for s in diag_strings[:100])
        raise PreprocessError(
            f"Unterminated string in expressions, line {current_start_line}. All strings: "
            + stringlist[:2000]
        )


def get_expressions(cont: str) -> list[tuple[str, int]]:
    tokens = tokenize(cont)
    return list(walk_tokens(tokens))



import argparse
import contextlib
import sys
from pathlib import Path

from rich.console import Console

from anyerplint.ast import Expression, FuncCall, SpecialToken
from anyerplint.preprocessor import PreprocessError, get_expressions
from anyerplint.tnex import ParseError, parse
from anyerplint.translate import Translator
from anyerplint.util import glob_plus, read_text_file

console = Console(highlight=False)


def emit(s: str) -> None:
    console.print(s)


def color(text: str, color: str) -> str:
    return f"[{color}]{text}[/{color}]"


def print_recursive_lists(tree: Expression, indent: int) -> None:
    if isinstance(tree, FuncCall):
        emit("  " * indent + color(tree.name, "deep_sky_blue1"))
        for arg in tree.args:
            print_recursive_lists(arg, indent + 1)
    else:
        el = tree
        if el == SpecialToken.EmptyArgument:
            col = None
            el = "<EMPTY>"
        elif el.startswith('"') and el.endswith('"'):
            col = "dark_orange"
        elif el.lower().startswith("v,"):
            col = "bright_green"
        elif el.isdecimal():
            col = "bright_yellow"
        else:
            col = None

        t = color(el, col) if col else el
        emit("  " * indent + t)


def color_translation_errors(expr: str) -> str:
    return expr.replace("~(~", "[red]").replace("~)~", "[/red]")


def search_calls(ast: FuncCall, func_name: str) -> list[FuncCall]:
    res = []
    func = func_name.upper()

    if ast.name.upper() == func:
        res.append(ast)

    for arg in ast.args:
        if not isinstance(arg, FuncCall):
            continue
        child_results = search_calls(arg, func_name=func_name)
        res.extend(child_results)

    return res


def pretty_flat_call(ast: FuncCall) -> str:
    ret = ast.name

    pretty_args = []
    for arg in ast.args:
        match arg:
            case str(a):
                pretty_args.append(repr(a))
            case FuncCall(name, args):
                pretty_args.append(f"{name}({len(args)})")

    return " ; ".join([ret, *pretty_args])


def search_and_print_calls(fname: Path, func_name: str) -> None:
    cont = read_text_file(fname)
    exps = get_expressions(cont)
    for exp in exps:
        parsed = parse(exp[0])
        if not isinstance(parsed, FuncCall):
            continue
        hits = search_calls(parsed, func_name=func_name)
        for hit in hits:
            printed = pretty_flat_call(hit)
            with contextlib.suppress(UnicodeEncodeError):
                emit(str(fname) + " ; " + printed)


def print_formatted_expression(
    exp: str, linenum: int, errors_only: bool, log_all: bool
) -> None:
    try:
        parsed = parse(exp)

        translator = Translator()
        translated = translator.translate(parsed)
    except (ValueError, ParseError) as e:
        emit(f"[red]Failed to parse[/]: {exp} ({e})")
        return

    if errors_only and not translator.errors:
        return

    non_trivial_expr = translated != exp
    if non_trivial_expr or log_all:
        emit(f"L[yellow]{linenum}[/]\t[grey66]{exp}[/]")
    if non_trivial_expr:
        emit("=>\t" + color_translation_errors(translated))
        expr = parsed
        if not isinstance(expr, FuncCall):
            return
        print_recursive_lists(expr, 1)
        emit("")


def print_all_expressions(cont: str, errors_only: bool, log_all: bool) -> None:
    try:
        expressions = get_expressions(cont)
    except PreprocessError as e:
        emit(f"[red]{e}[/red]")
    else:
        for exp, linenum in expressions:
            print_formatted_expression(exp, linenum, errors_only, log_all)


def handle_parse(args: argparse.Namespace) -> None:
    errors_only = args.errors
    console.no_color = args.no_color
    if args.fun:
        if not args.filename:
            emit("With 'parse --fun' mode, you must specify file or directory to check")
            return
        for pat in args.filename:
            fnames = glob_plus(pat)
            for f in fnames:
                with contextlib.suppress(
                    PreprocessError, ParseError, UnicodeDecodeError, PermissionError
                ):
                    search_and_print_calls(f, args.fun)

        return

    if not args.filename:
        console.rule("stdin")
        cont = sys.stdin.read()
        print_all_expressions(cont, errors_only, log_all=True)

    for pat in args.filename:
        fnames = glob_plus(pat)
        for f in fnames:
            console.rule(str(f))
            cont = read_text_file(f)
            print_all_expressions(cont, errors_only, log_all=False)


def init_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Declare arguments you need here."""
    parser.add_argument("filename", nargs="*", help="Files to parse")
    parser.add_argument(
        "--errors", action="store_true", help="Print only expressions with errors"
    )
    parser.add_argument("--fun", help="Only print certain functions, e.g. --fun F,EVAL")
    parser.add_argument(
        "--no-color",
        action="store_true",
        help="No colors, good for redirecting to file",
    )
    parser.set_defaults(func=handle_parse)
    return parser



import argparse
import sys

from anyerplint import business_logic


def handle_check(args: argparse.Namespace) -> None:
    """Implement argument handling here.

    Don't put business logic here. Only parse the arguments and pass forward.
    """
    errors = business_logic.do_check(args.lib, args.target, args.teach)
    if args.exitcode and errors:
        sys.exit(1)


def init_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Declare arguments you need here."""
    parser.add_argument(
        "-l",
        "--lib",
        action="append",
        default=[],
        help="Library path containing variable declarations",
    )

    parser.add_argument("target", nargs="+", help="Target file or directory to check")

    parser.add_argument(
        "--teach",
        action="store_true",
        help="Run in teaching mode (update function list)",
    )

    parser.add_argument(
        "--exitcode", action="store_true", help="Set nonzero exit code on error"
    )

    parser.set_defaults(func=handle_check)
    return parser



import glob
import itertools
import os
import re
from collections.abc import Iterable, Iterator
from pathlib import Path
from typing import TypeVar
from xml.etree.ElementTree import Element

from charset_normalizer import from_path

T = TypeVar("T")


def count_trailing(source: str, chars: str) -> int:
    return len(source) - len(source.rstrip(chars))


def skip(iterable: Iterable[T], count: int) -> Iterator[T]:
    return itertools.islice(iterable, count, None)


def glob_plus(pattern: str, extension: str = "xml") -> list[Path]:
    if "*" in pattern:
        # Path.glob doesn't seem to support absolute paths
        return [Path(p) for p in glob.glob(pattern, recursive=True)]  # noqa: PTH207

    path = Path(pattern)
    if path.is_dir():
        return list(path.glob(f"**/*.{extension}"))

    return [path]


XML_ENCODING_DECLARATION = re.compile(rb'encoding="(.*?)"', re.DOTALL)


# This tries to detect the encoding from the XML declaration,
# and then reads the file with that encoding.
# This only checks for ASCII-compatible encodings, UTF-16 etc. are not supported.
def _read_text_file_fast(file_path: Path) -> str | None:
    try:
        with file_path.open("rb") as file:
            data = file.read(50)
        match = XML_ENCODING_DECLARATION.search(data)
        if match:
            # The encoding should be ASCII-compatible, since it matched the binary regex.
            enc = match.group(1).decode()
            return file_path.read_text(encoding=enc)
    except UnicodeDecodeError:
        pass

    return None


def _read_text_file_slow(file_path: Path) -> str:
    matches = from_path(file_path)
    best = matches.best()
    return str(best)


def read_text_file(file_path: Path) -> str:
    if file_path.suffix == ".xml":
        from_decl = _read_text_file_fast(file_path)
        if from_decl:
            return from_decl

    return _read_text_file_slow(file_path)


def format_xml_tag(elem: Element) -> str:
    if not elem.attrib:
        return f"<{elem.tag}>"
    attributes = " ".join(f'{name}="{value}"' for name, value in elem.attrib.items())
    return f"<{elem.tag} {attributes}>"


def get_app_local_dir() -> Path:
    local_app_data = os.getenv("LOCALAPPDATA")
    if local_app_data:
        return Path(local_app_data) / "anyerplint"
    msg = "Could not find LOCALAPPDATA"
    raise Exception(msg)


def add_linenumbers(cont: str, needles: list[str]) -> list[str]:
    hits: list[list[int]] = [[] for i in range(len(needles))]
    for linenum, line in enumerate(cont.splitlines(), 1):
        for _in, n in enumerate(needles):
            if n in line:
                hits[_in].append(linenum)

    return [
        n + " - line " + ", ".join(map(str, hits[idx]))
        for (idx, n) in enumerate(needles)
    ]


def _replace_with_empty(match: re.Match[str]) -> str:
    comment = match.group(0)
    return "\n" * comment.count("\n")


def replace_commented_xml_with_empty_lines(xml_string: str) -> str:
    comment_pattern = "<!--(.*?)-->"
    return re.sub(comment_pattern, _replace_with_empty, xml_string, flags=re.DOTALL)


def replace_cdata_with_empty_lines(xml_string: str) -> str:
    cdata_pattern = r"<!\[CDATA\[(.*?)\]\]>"
    return re.sub(cdata_pattern, _replace_with_empty, xml_string, flags=re.DOTALL)


XML_ENTITY_PATTERN = re.compile(r"&([^;]+);")


def expand_xml_entities(xml_string: str) -> str:
    def replace_entity(match: re.Match[str]) -> str:
        entity = match.group(1)
        if entity == "lt":
            return "<"
        elif entity == "gt":
            return ">"
        elif entity == "amp":
            return "&"
        elif entity == "quot":
            return '"'
        else:
            return match.group(0)

    return XML_ENTITY_PATTERN.sub(replace_entity, xml_string)



from anyerplint.preprocessor import get_expressions

test_cont = r"""1 filler
2 lines
3 Hello, {v,fooo}
4 {oeoeoe}
5
6 Hello {v,foo("}}")}
7
8 More nesting {oeoe}
9 {"string"}
10 {"st with \" and other stuff"}
11 {
12 many
13 lines
14 }
"""


def test_preproc() -> None:
    expressions = get_expressions(test_cont)
    assert expressions == [
        ("v,fooo", 3),
        ("oeoeoe", 4),
        ('v,foo("}}")', 6),
        ("oeoe", 8),
        ('"string"', 9),
        ('"st with \\" and other stuff"', 10),
        ("\n12 many\n13 lines\n14 ", 11),
    ]



from dataclasses import dataclass
from enum import Enum, auto
from typing import Union


class SpecialToken(Enum):
    """Special tokens that can appear in the AST."""

    EmptyArgument = auto()


Expression = Union[str, "FuncCall", SpecialToken]


@dataclass
class FuncCall:
    """An AST node representing a function call."""

    name: str
    args: list[Expression]



from pathlib import Path

import pytest
from anyerplint import business_logic

root = Path(__file__).absolute().parent


def test_brace_check() -> None:
    tested = business_logic.brace_check("(({[]}))")
    assert tested == []
    tested = business_logic.brace_check("[")
    [err] = tested
    assert err.startswith("File ended with mismatched braces")


def test_check_minimal_without_lib() -> None:
    minimal_script = root / "minimal_script.xml"
    res = business_logic.do_check(
        [], [str(minimal_script)], teaching=False, nostdlib=True
    )
    expected = {
        str(minimal_script): {
            "Unknown tags": [
                "<erpConnector>",
                "<error>",
                "<section name>",
                "<sections>",
            ],
            "Other errors": [
                "File ended with mismatched braces, remaining in stack (char, linenum): ( 5",
            ],
        },
    }
    assert res == expected


def test_check_malformed() -> None:
    malformed_file = root / "malformed.xml"
    res = business_logic.do_check(
        [], [str(malformed_file)], teaching=False, nostdlib=True
    )
    expected = {
        str(
            malformed_file,
        ): "FATAL, not well-formed (invalid token): line 4, column 19",
    }
    assert res == expected


@pytest.mark.skip()
def test_check_expression_like_regexes() -> None:
    input_file = root / "regexes.xml"
    res = business_logic.do_check([], [str(input_file)], teaching=False, nostdlib=True)
    expected = {
        "Unknown tags": [
            "<erpConnector>",
            "<section name>",
            "<sections>",
            "<variable name>",
        ],
    }

    errs = res.get(str(input_file), {})
    assert isinstance(errs, dict)
    assert not errs.get("Other errors")
    assert errs == expected


def test_check_expressions() -> None:
    input_file = root / "expressions.xml"
    res = business_logic.do_check([], [str(input_file)], teaching=False, nostdlib=True)
    expected = {
        "Unknown tags": [
            "<erpConnector>",
            "<section name>",
            "<sections>",
            "<variable name>",
        ],
        "Expression errors": [
            {
                "Error": "Line 5, expression contains no comma: {nocomma}",
                "Simple": "N/A",
                "Msg": [],
            }
        ],
    }

    errs = res.get(str(input_file), {})
    assert isinstance(errs, dict)
    assert not errs.get("Other errors")
    assert errs == expected



import pytest
from anyerplint.ast import FuncCall, SpecialToken
from anyerplint.tnex import ParseError, parse


def test_parser() -> None:
    assert parse("foo") == "foo"
    assert parse("foo()") == FuncCall("foo", [])
    assert parse("foo(),N") == FuncCall(",N", [FuncCall("foo", [])])
    assert parse("foo ()") == FuncCall("foo", [])
    assert parse("foo(1)") == FuncCall("foo", ["1"])
    assert parse("foo(1;1)") == FuncCall("foo", ["1", "1"])

    assert parse("foo(1;;1)") == FuncCall("foo", ["1", SpecialToken.EmptyArgument, "1"])
    assert parse("foo(1; 1)") == FuncCall("foo", ["1", "1"])
    assert parse("foo(1; ;1)") == FuncCall(
        "foo", ["1", SpecialToken.EmptyArgument, "1"]
    )
    assert parse('foo(1; "1")') == FuncCall("foo", ["1", '"1"'])
    assert parse("foo(1;)") == FuncCall("foo", ["1", SpecialToken.EmptyArgument])
    assert parse("foo(;1)") == FuncCall("foo", [SpecialToken.EmptyArgument, "1"])
    assert parse("foo(1; )") == FuncCall("foo", ["1", SpecialToken.EmptyArgument])
    assert parse('foo(1;;"1")') == FuncCall(
        "foo", ["1", SpecialToken.EmptyArgument, '"1"']
    )
    assert parse('foo(1;; "1")') == FuncCall(
        "foo", ["1", SpecialToken.EmptyArgument, '"1"']
    )

    assert parse('"some string"') == '"some string"'

    assert parse('foo("some string";"other string";0)') == FuncCall(
        "foo", ['"some string"', '"other string"', "0"]
    )

    assert parse('foo("some string";"other string")') == FuncCall(
        "foo", ['"some string"', '"other string"']
    )

    # nothing special with ,
    assert parse("a,b(c,d;e,f;1.0)") == FuncCall("a,b", ["c,d", "e,f", "1.0"])

    assert parse(r'" string with \" char"') == '" string with \\" char"'
    assert parse(R'" string with \\\" char"') == '" string with \\\\\\" char"'
    assert (
        parse(r'"some string with ) and ( and \" characters"')
        == '"some string with ) and ( and \\" characters"'
    )

    assert parse(r'foo,"long / () accessor"') == 'foo,"long / () accessor"'
    assert parse(r'foo(F,NOW(),"yyyy")') == FuncCall(
        "foo", [FuncCall(',"yyyy"', [FuncCall("F,NOW", [])])]
    )
    assert parse(r'F,NOW(),"yyyy"') == FuncCall(',"yyyy"', [FuncCall("F,NOW", [])])

    complex_expr = parse(
        r'foo(f1;bar(b1;b2;"b3";"wide string");f2;baz(1;"some string with ) and ( and \" characters"))'
    )
    assert complex_expr == FuncCall(
        "foo",
        [
            "f1",
            FuncCall("bar", ["b1", "b2", '"b3"', '"wide string"']),
            "f2",
            FuncCall("baz", ["1", '"some string with ) and ( and \\" characters"']),
        ],
    )

    with pytest.raises(ParseError, match="Unterminated string"):
        parse('foo("unterminated)')

    with pytest.raises(
        ParseError, match="Invalid token ';' at top level of expression"
    ):
        parse(";")

    with pytest.raises(
        ParseError, match=r"Invalid token '\)' at top level of expression"
    ):
        parse("foo)")

    with pytest.raises(ParseError, match="Empty parse result for expression: ''"):
        parse("")

    with pytest.raises(
        ParseError,
        match="',format' syntax called without preceding expression: ',hello'",
    ):
        parse(",hello")

    with pytest.raises(
        ParseError,
        match="',format' syntax called without preceding expression: ',\"hello\"'",
    ):
        parse(',"hello"')



"""tnex - Trivial Nested EXpressions.

Easier to read/write than s-expr's I guess

"""

import re

from anyerplint import util
from anyerplint.ast import Expression, FuncCall, SpecialToken
from anyerplint.util import expand_xml_entities


def tokenize(s: str) -> list[str]:
    # split to parts separated by ; ( ) "
    tokens = re.split(r"([\(\)\";])", s)
    filtered = list(filter(None, tokens))
    return _combine_tokens(filtered)


def _combine_tokens(tokens: list[str]) -> list[str]:
    result = []
    i = 0
    while i < len(tokens):
        tok = tokens[i]
        if tok == '"':
            s, moved = _parse_string(tokens[i:])
            result.append(s)
            i += moved
        else:
            stripped = tok.strip()
            if stripped:
                result.append(stripped)
            i += 1
    return result


class ParseError(Exception):
    """Exceptions raised by parser."""


def _parse_string(toks: list[str]) -> tuple[str, int]:
    assert toks[0] == '"'
    # eat up tokens to produce just one str
    result = ['"']
    escape = False
    for tok in util.skip(toks, 1):
        result.append(tok)
        if tok.endswith("\\"):
            backslashes = util.count_trailing(tok, "\\")
            escape = (backslashes % 2) != 0
        elif tok == '"' and not escape:
            value = "".join(result)
            return value, len(result)
        else:
            escape = False

    msg = "Unterminated string"
    raise ParseError(msg)


def _parse_accessor(first: str, toks: list[str]) -> tuple[str, int]:
    if toks and toks[0].startswith('"'):
        return first + toks[0], 2

    return first, 1


def emit_nested_sequence(
    parts: list[str], top_level: bool
) -> tuple[list[Expression], int]:
    res: list[Expression] = []
    add_empty_arg = True
    i = 0
    while i < len(parts):
        match parts[i:]:
            case [";", *_]:
                if top_level:
                    msg = "Invalid token ';' at top level of expression"
                    raise ParseError(msg)
                if add_empty_arg:
                    res.append(SpecialToken.EmptyArgument)
                add_empty_arg = True
                i += 1
            case [")", *_]:
                if top_level:
                    msg = "Invalid token ')' at top level of expression"
                    raise ParseError(msg)
                if add_empty_arg and res:
                    res.append(SpecialToken.EmptyArgument)
                i += 1
                break
            case [func_name, "(", *rest]:
                nested, moved = emit_nested_sequence(rest, top_level=False)
                func = FuncCall(name=func_name, args=nested)
                res.append(func)
                add_empty_arg = False
                i += moved + 1
            case [it, *rest] if it.startswith(","):
                # actually call previous output with "nesting" output
                s, moved = _parse_accessor(it, rest)
                try:
                    previous = res.pop()
                except IndexError as exc:
                    msg = f"',format' syntax called without preceding expression: '{s}'"
                    raise ParseError(msg) from exc

                # Not actually a function call
                fmt = FuncCall(s, [previous])
                add_empty_arg = False
                res.append(fmt)
                i += moved

            # special foo,"hello" accessor that accesses property of foo.
            # lexer breaks it because of " char, so reassemble it here
            case [it, *rest] if it.endswith(","):
                s, moved = _parse_accessor(it, rest)
                add_empty_arg = False
                res.append(s)
                i += moved
            case [it, *_]:
                add_empty_arg = False
                res.append(it.strip())
                i += 1

    return (res, i + 1)


def parse(s: str, *, expand_entities: bool = True) -> Expression:
    if expand_entities:
        s = expand_xml_entities(s)
    tokens = tokenize(s)
    parsed, _ = emit_nested_sequence(tokens, top_level=True)
    if not parsed:
        msg = f"Empty parse result for expression: '{s}'"
        raise ParseError(msg)

    return parsed[0]



from anyerplint.translate import translate_str


def test_translate_syntax() -> None:
    assert (
        translate_str(
            'F,EVAL(obj.A,AssetCode;=;"";F,EVAL(obj.A,SUM;"&gt;";"0";40;50);F,EVAL(obj.A,SUM;"&gt;";"0";70;75))'
        )
        == '(obj.A,AssetCode = "" ? (obj.A,SUM > "0" ? 40 : 50) : (obj.A,SUM > "0" ? 70 : 75))'
    )

    assert (
        translate_str(
            "F,EVAL(F,EXISTS(v;config);=;1;F,REPLACE(F,LOWER(v,config);rootconfig;systemvars);&quot;&quot;)"
        )
        == '(defined(config) = 1 ? v,config.lower().replace(rootconfig -> systemvars) : "")'
    )

    assert translate_str("F,NOT(v,x)") == "not v,x"



import argparse
import shutil
import urllib.error
import urllib.request
from pathlib import Path

from anyerplint import util


def handle_import(args: argparse.Namespace) -> None:
    """Implement argument handling here.

    Don't put business logic here. Only parse the arguments and pass forward.
    """
    do_import(args.target)


def init_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Declare arguments you need here."""
    parser.add_argument("target", nargs="+", help="Zip file to import definitions from")

    parser.set_defaults(func=handle_import)
    return parser


emit = print


def do_import(fnames: list[str]) -> None:
    target_dir = util.get_app_local_dir()
    if not target_dir.is_dir():
        target_dir.mkdir(parents=True)
    for f in fnames:
        if f.startswith(("http:", "https:")):
            lastpart = f.split("/")[-1]
            emit(lastpart)

            if not lastpart.endswith(".zip"):
                lastpart = "downloaded.zip"
            tfile = Path(target_dir) / lastpart
            emit("Fething:", f, "->", tfile)
            try:
                urllib.request.urlretrieve(f, tfile)  # noqa: S310
            except urllib.error.URLError:
                emit(
                    "Failed to download, ensure your VPN is operational and the target file exists!"
                )
        elif f.endswith(".zip"):
            emit("Copying:", f, "->", target_dir)
            shutil.copy(f, target_dir)
        else:
            emit("Not a zip file:", f)



from anyerplint.ast import Expression, SpecialToken
from anyerplint.tnex import ParseError, parse


def removequote(s: str) -> str:
    return s.removeprefix('"').removesuffix('"')


def translate_str(s: str) -> str | list[str]:
    parsed = parse(s)
    translator = Translator()
    result = translator.translate(parsed)
    errors = translator.errors
    return errors if errors else result


# these will be converted to uppercase variants. Eventually this should be empty but needs fixing a lot of templates

allow_wrong_case = {
    "F,Calc",
    "F,Combine",
    "F,Eval",
    "F,IsChanged",
    "F,Not",
    "F,SetData",
    "F,ToDate",
}

# these will be translated to name(arg1; arg2)
nnary_funcs: dict[str, int | tuple[int, int]] = {
    "F,ACCEPTCHANGES": (1, 2),
    "F,ADDXML": (3, 4),
    "F,ADDXMLNS": 3,
    "F,BASE64DECODE": 1,
    "F,BASE64ENCODE": 1,
    "F,CLEARERROR": 0,
    "F,CSV": (2, 3),
    "F,DATEADD": 3,
    "F,DATEDIFF": 3,
    "F,DBSELECT": (2, 999),
    "F,DECRYPT": (1, 2),
    "F,ENCRYPT": (1, 2),
    "F,FLOOR": 1,
    "F,GETCREDENTIALS": (2, 3),
    "F,GUID": 0,
    "F,ISCHANGED": 2,
    "F,ISDATASOURCECHANGED": 1,
    "F,ISDATE": (1, 2),
    "F,ISNOTHING": 1,
    "F,ISNUMERIC": 1,
    "F,LPAD": 3,
    "F,NOW": 0,
    "F,PARSE": 1,
    "F,RANGE": 4,
    "F,REPLICATE": 2,
    "F,REVERSE": 1,
    "F,ROUND": (2, 3),
    "F,RPAD": 3,
    "F,SAPSELECT": (6, 12),
    "F,SAPSELECTWS": (7, 12),
    "F,XMLENCODE": 1,
}


class Translator:
    """Translates expressions into a simpler form."""

    def __init__(self) -> None:
        """Initialize a translator with empty error list."""
        self.errors: list[str] = []

    def translate(self, tree: Expression) -> str:
        """Convert to pretty mostly-infix notation."""
        # only function calls are translated
        if tree == SpecialToken.EmptyArgument:
            return ""
        if isinstance(tree, str):
            return tree

        translate = self.translate
        func_name = tree.name
        if not func_name.isupper() and func_name.upper().startswith("F,"):
            if func_name not in allow_wrong_case:
                self.errors.append(f"Function name {func_name} should be uppercase")
            func_name = func_name.upper()

        match (func_name, tree.args):
            case "F,EVAL", [obj, operation, comp, iftrue, iffalse]:
                return f"({translate(obj)} {removequote(translate(operation))} {translate(comp)} ? {translate(iftrue)} : {translate(iffalse)})"
            case "F,EXISTS", ["v", key]:
                return f"defined({key})"
            case "F,EXISTS", [obj, key]:
                return f"({key} in {translate(obj)})"
            case "F,EXISTS", [obj]:
                return f"(exists {translate(obj)})"

            case "F,REPLACE", [src, frome, toe]:
                return (
                    f"{translate(src)}.replace({translate(frome)} -> {translate(toe)})"
                )
            case "F,LOWER", [exp]:
                return f"{translate(exp)}.lower()"
            case "F,UPPER", [exp]:
                return f"{translate(exp)}.upper()"
            case "F,TRIM", [exp]:
                return f"{translate(exp)}.trim()"
            case "F,LTRIM", [exp]:
                return f"{translate(exp)}.ltrim()"
            case "F,RTRIM", [exp]:
                return f"{translate(exp)}.rtrim()"

            case "F,NVL", [exp, default]:
                return "(" + translate(exp) + " ?? " + translate(default) + ")"

            case "F,TONUMBER", [exp, '"."']:
                return f"num({translate(exp)})"
            case "F,TONUMBER", [exp, sep]:
                return f"num({translate(exp)} - {translate(sep)})"
            case "F,TOCHAR", [exp, str(format)]:
                return f"num({translate(exp)}.tochar({format})"
            case "F,TOCHAR", [exp, str(format), str(culture)]:
                return f"num({translate(exp)}.tochar({format}, {culture})"
            case "F,FORMAT", [exp, format]:
                return f"{translate(exp)}.format({translate(format)})"
            case "F,COMBINE", parts:
                translated = [translate(part) for part in parts]
                return "(" + " & ".join(translated) + ")"
            case "F,GETDATA", [ds, key]:
                return f"{translate(ds)}[|{translate(key)}|]"
            case "F,GETNODE", [src, path]:
                return f"{translate(src)}.node({translate(path)})"
            case "F,SETDATA", [ds, key, value]:
                return f"{translate(ds)}[|{translate(key)}|] := {translate(value)}"
            case "F,IF", [src, op, tgt]:
                return (
                    f"if {translate(src)} {removequote(translate(op))} {translate(tgt)}"
                )
            case "F,LEN", [o]:
                return f"len({translate(o)})"
            case "F,CALC", parts:
                return "(" + " ".join(translate(part) for part in parts) + ")"
            case "F,NOT", [exp]:
                return "not " + translate(exp)
            case "F,AND", conds:
                translated = [translate(part) for part in conds]
                return "(" + " && ".join(translated) + ")"
            case "F,OR", [*conds]:
                translated = [translate(part) for part in conds]
                return "(" + " || ".join(translated) + ")"
            case "F,ROWCOUNT", [o]:
                return f"{translate(o)}.rowcount()"
            case "F,TODATE", [d, str(format)]:
                return f"{translate(d)}.todate({format})"
            case "F,CHR", [str(code)] if code.isdigit():
                return f"char '{chr(int(code))}' {code}"
            case "F,CURSORINDEX", [str(src)]:
                return f"{src}.cursorindex()"
            case "F,SUBSTR", [exp, beg, count]:
                return f"{translate(exp)}.substr({translate(beg)}, {translate(count)})"
            case "F,INSTR", [haystack, needle, then, else_]:
                return f"{translate(needle)} in_string {translate(haystack)} ? {translate(then)} : {translate(else_)}"
            case "F,RIGHT", [exp, count]:
                return f"{translate(exp)}[-{translate(count)}:]"
            case "F,LEFT", [exp, count]:
                return f"{translate(exp)}[:{translate(count)}]"
            case "F,REGEXP_STRING", [s, pat]:
                return f"{translate(s)}.re_search({translate(pat)})"
            case "F,REGEXP_POSITION", [s, pat]:
                return f"{translate(s)}.re_search_pos({translate(pat)})"
            case str(func), [param] if func.startswith(","):
                return f"{translate(param)}.pipe({func})"
            case str(func), args if func in nnary_funcs:
                match nnary_funcs[func]:
                    case int(n):
                        arity = (n, n)
                    case (int(begg), int(endd)):
                        arity = (begg, endd)

                if not arity[0] <= len(args) <= arity[1]:
                    self.errors.append(
                        f"Invalid number of arguments for {func} - got {len(args)}, expected {nnary_funcs[func]}"
                    )
                name = f"{func.removeprefix('F,')}".lower()
                joined = "; ".join(translate(part) for part in args)
                return f"{name}({joined})"

            case str(func), args:
                newname = func.removeprefix("F,")
                self.errors.append(f"Unknown function {func} ({len(args)} arguments)")
                return f"~(~{newname}(" + ";".join(translate(r) for r in args) + ")~)~"

        s = f"Unmatched pattern {tree}"
        raise ParseError(s)
